/*
 * Copyright (C) 2013-2018 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <errno.h>
#include <string.h>

#include "mockturtle-rt.h"
#include <mockturtle-framework.h>
#include "wrtd-common.h"

#define OUT_QUEUE_MAXTIME 10
#define OUT_QUEUE_PREFIX adcin_
#define OUT_QUEUE_SIZE 4

#include "out_queue.h"

struct wrtd_adcin_dev {
        uint32_t io_addr;
		adcin_out_queue queue;
};

static inline int adcin_wr_present(struct wrtd_adcin_dev *fd)
{
	return 1;
}

static inline int adcin_wr_link_up(struct wrtd_adcin_dev *fd)
{
	return 1;
}

static inline int adcin_wr_time_locked(struct wrtd_adcin_dev *fd)
{
	return 1;
}

static inline int adcin_wr_time_ready(struct wrtd_adcin_dev *fd)
{
	return 1;
}

static inline int adcin_wr_sync_timeout(void)
{
	return 0;
}

static void adcin_wr_enable_lock(struct wrtd_adcin_dev *fd, int enable)
{
	return;
}


static inline void adcin_writel(struct wrtd_adcin_dev *dev, uint32_t value,
				 uint32_t reg)
{
	dp_writel(value, dev->io_addr + reg);
}

static inline uint32_t adcin_readl (struct wrtd_adcin *dev, uint32_t reg)
{
	return dp_readl(dev->io_addr + reg);
}



/**
 * Drop the given enqueued trigger
 */
static void drop_trigger(struct wrtd_fd_channel *out,
			 struct wrtd_event *ev,
			 struct lrt_pulse_queue *q, unsigned reason)
{
	out->idle = 1;

	if (pulse_queue_empty(q))
		return;

	/* Drop the pulse */
	pulse_queue_pop(q);

	/* Disarm the FD output */
	fd_ch_writel(out, FD_DCR_MODE, FD_REG_DCR);

	wrtd_log(WRTD_LOG_MSG_EV_DISCARDED, reason, ev, NULL);
}


/**
 * Output driving function. Reads pulses from the output queue,
 * programs the output and updates the output statistics.
 */
static void adcin_output (struct wrtd_adcin_dev *dev)
{
	struct adcin_out_queue *q = &dev->queue;
	struct wrtd_event *ev = adcin_out_queue_front(q);
	uint32_t ctrl = adcin_readl(dev, ALT_TRIGIN_CTRL);
	struct wrtd_tstamp *ts;

	/* Check if the output has triggered */
	if (!dev->idle) {
#if 0
		if (!wr_is_timing_ok()) {
			/* Timing has been lost.  */
			drop_trigger(out, ev, q, WRTD_LOG_DISCARD_NO_SYNC);
			return;
		}
#endif
		if (ctrl & ALT_TRIGIN_CTRL_ENABLE) {
			/* Armed but still waiting for trigger */
			if (check_output_timeout (dev)) {
				/* Will never trigger.  Missed.  */
				drop_trigger(dev, ev, q, WRTD_LOG_DISCARD_TIMEOUT);
			}
		} else {
			/* Has been triggered.  */
			wrtd_log(WRTD_LOG_MSG_EV_CONSUMED, WRTD_LOG_CONSUMED_DONE, ev, NULL);

			adcin_out_queue_pop(q);
			dev->idle = 1;
		}
		return;
	}

	/* Output is idle: check if there's something in the queue to execute */
	if (adcin_out_queue_empty(q))
		return;

	ev = pulse_queue_front(q);
	ts = &ev->ts;

	if (!wr_is_timing_ok()) {
		drop_trigger(out, ev, q, WRTD_LOG_DISCARD_NO_SYNC);
		return;
	}

	/* Program the output start time */
	adcin_writel(dev, ts->seconds, ALT_TRIGIN_SECONDS + 0);
	adcin_writel(dev, ts->ns / 8, ALT_TRIGIN_CYCLES);
	adcin_writel(dev, ALT_TRIGIN_CTRL_ENABLE, ALT_TRIGIN_CTRL);

	wrtd_log(WRTD_LOG_MSG_EV_CONSUMED, WRTD_LOG_CONSUMED_START, ev, NULL);

	ts_add2_ns (ts, 8000);

	/*
	 * Store the last programmed timestamp (+ some margin) and mark
	 * the output as busy
	 */
	q->last_programmed_sec = ts->seconds;
	q->last_programmed_ns = ts->ns;
	dev->idle = 0;
}

static void fd_local_output(struct wrtd_fd_dev *fd,
							struct wrtd_event *ev, unsigned ch)
{
	struct wrtd_fd_channel *out = &fd->channels[ch];
	struct wrtd_event *pq_ev;

	pq_ev = pulse_queue_push(&out->queue);
	if (!pq_ev) {
				/* overflow.
				   FIXME: stats ?  */
				wrtd_log(WRTD_LOG_MSG_EV_DISCARDED, WRTD_LOG_DISCARD_OVERFLOW,
						 ev, NULL);
		return;
	}

	*pq_ev = *ev;
}

#if 0
/**
 * It disable the given channel and clear its internal queue
 */
static int wrtd_out_disable(struct trtl_fw_msg *msg_i,
				struct trtl_fw_msg *msg_o)
{
	uint32_t ch = ((uint32_t *)msg_i->payload)[0];

	pulse_queue_init(&wrtd_out_channels[ch].queue);
	fd_ch_writel(&wrtd_out_channels[ch], FD_DCR_MODE, FD_REG_DCR);

	msg_o->header->msg_id = msg_i->header->msg_id;

	return 0;
}
#endif

static void wrtd_adcin_data_init(struct wrtd_adcin_dev *dev)
{
	adcin_out_queue_init(&fd->channels[i].queue);
	adcin->idle = 1;
}
